<?php
/*
 * Copyright (c) 2012, webvariants GbR, http://www.webvariants.de
 *
 * This file is released under the terms of the MIT license. You can find the
 * complete text in the attached LICENSE file or online at:
 *
 * http://www.opensource.org/licenses/mit-license.php
 */

class WV2_RestrictArticletype_ArticleTypesDatatype extends _WV_Datatype_Select {
	public function deserializeValue($value, array $params) {
		if (empty($value)) return array();

		$values = array_map('strval', explode('|', $value));
		$values = array_filter($values);

		return $values;
	}

	public function serializeValue($value, array $params) {
		return is_array($value) && !empty($value) ? '|'.implode('|', $value).'|' : '';
	}

	public function checkParamCompatibility(array $oldParameters, array $newParameters) {
		return true;
	}

	public function getIncompatibilityUpdateStatement(array $oldParameters, array $newParameters) {
		return array();
	}

	protected function getDefaultParams() {
		$service = sly_Service_Factory::getArticleTypeService();
		$types   = array_keys($service->getArticleTypes());

		return array(
			'inheit'       => false,
			'types'        => $types,
			'sourceValues' => $types,
		);
	}

	protected function deserializeParams(array $params, $prefix = '') {
		return _WV_Datatype_Base::deserializeParams($params, $prefix);
	}

	protected function decodeFormValue(WV_IProperty $property, $value) {
		$value  = _WV_Datatype_Base::decodeFormValue($property, $value);
		$params = $property->getParams();

		if ($property->isMultilingual()) {
			foreach ($value as $clang => $list) {
				$value[$clang] = $this->deserializeValue($list, $params);
			}
		}
		else {
			$value = $this->deserializeValue($value, $params);
		}

		return $value;
	}

	public function renderForm(sly_Form $form, WV_IProperty $property, $value) {
		$value  = $this->decodeFormValue($property, $value);
		$params = $property->getParams();

		extract($this->deserializeParams($params));

		if ($inherit) {
			if (sly_Core::getCurrentPage() !== 'structure') {
				throw new WV2_Exception('Inherited article types work only with category metainfos.');
			}

			$catID = sly_request('category_id', 'int');
			$cat   = sly_Util_Category::findById($catID); // it's already the parent of the currently edited category
			$all   = $sourceValues;

			if ($cat) {
				$allowed      = WV2_RestrictArticletype_Helper::getAllowedTypes($cat);
				$sourceValues = array_intersect($sourceValues, $allowed);
			}
		}

		$service = sly_Service_Factory::getArticleTypeService();

		foreach ($sourceValues as $name) {
			$values[$name] = $service->getTitle($name);
		}

		$propID = $property->getID();
		$title  = $property->getTitle();
		$id     = 'property_'.$propID.'_value';

		if (count($values) === 1) {
			$select = new sly_Form_Text($title, reset($values));
		}
		else {
			$select = new sly_Form_Select_DropDown($id, $title, $value, $values);

			$select->setSize(min(10, count($sourceValues)));
			$select->setMultiple(true);
			$select->setHelpText($property->getHelpText());
		}

		$select->setMultilingual($property->isMultilingual());
		$form->add($select);

		return $form;
	}

	public function serializeFormFilter($list, WV_IProperty $property) {
		$params = $property->getParams();
		extract($this->deserializeParams($params));

		if ($inherit) {
			if (sly_Core::getCurrentPage() !== 'structure') {
				throw new WV_Exception_Datatype('Inherited article types work only with category metainfos.');
			}

			$catID = sly_request('category_id', 'int');
			$cat   = sly_Util_Category::findById($catID); // it's already the parent of the currently edited category
			$self  = sly_request('edit_id', 'int');
			$all   = $sourceValues;

			if ($cat) {
				$sourceValues = WV2_RestrictArticletype_Helper::getAllowedTypes($cat);
			}
		}

		foreach ($list as $clangID => $values) {
			$values = sly_makeArray($values);
			$result = array();

			// Gültige Elemente auswählen

			foreach ($values as $element) {
				if (in_array(strval($element), $sourceValues, true)) $result[] = $element;
			}

			// if every type was selected it's the same as if there was no selection at all

			if (count($result) === count($sourceValues)) {
				$result = array();
			}

			// if there has been *any* restriction, check if there are any articles
			// within this category that would violate this restriction

			if (!empty($result)) {
				$filter = new WV_Filter_And();
				$filter->add(new WV_Filter_Article_CategoryID($self, '=', true));

				$or = new WV_Filter_Or();
				$or->add(new WV_Filter_Article_Type(''));

				foreach ($result as $type) {
					$or->add(new WV_Filter_Article_Type($type));
				}

				$filter->add(new WV_Filter_Not($or));

				$articles = WV2_MetaProvider::filterArticles($filter);

				if (!empty($articles)) {
					$types   = array();
					$service = sly_Service_Factory::getArticleTypeService();

					foreach ($articles as $art) {
						$types[] = sly_translate($service->getTitle($art->getType()));
					}

					$types = array_unique($types);

					throw new WV_Exception_Datatype('Es gibt noch Artikel innerhalb dieser Kategorie, die nicht den neuen Beschränkungen des Artikeltyps genügen ('.implode(', ', $types).').');
				}

				// remember that we have reduced the number of possible articletypes
				// so that we can update children categories in WV2_CATEGORY_META_UPDATED.

				$listener = array('WV2_RestrictArticletype_Events', 'updateCategories');
				$params   = array('types' => $result, 'forbidden' => array_diff($all, $result));

				sly_Core::dispatcher()->register('WV2_CATEGORY_META_UPDATED', $listener, $params);
			}

			$list[$clangID] = empty($result) ? '' : '|'.implode('|', $result).'|';
		}

		return $list;
	}
}
